/* SPIM S20 MIPS simulator.
   Misc. routines for SPIM.
   Copyright (C) 1990 by James Larus (larus@cs.wisc.edu).

   SPIM is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 1, or (at your option) any
   later version.

   SPIM is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License
   along with GNU CC; see the file COPYING.  If not, write to James R.
   Larus, Computer Sciences Department, University of Wisconsin--Madison,
   1210 West Dayton Street, Madison, WI 53706, USA or to the Free
   Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */


/* $Header: /var/home/larus/SPIM/RCS/spim-utils.c,v 3.14 1991/05/09 21:00:44 larus Exp larus $
*/


#include <stdio.h>
#include <varargs.h>
#include <ctype.h>

#include "spim.h"
#include "inst.h"
#include "mem.h"
#include "reg.h"
#include "y.tab.h"


/* Exported functions: */

void add_breakpoint ();
void delete_breakpoint ();
void fatal_error ();
void initialize_registers ();
void initialize_run_stack ();
void initialize_world ();
void list_breakpoints ();
int read_assembly_file ();
int run_program ();
mem_addr starting_address ();

/* Imported functions: */

void data_begins_at_point ();
void exit ();
void flush_local_labels ();
void initialize_scanner ();
void initialize_symbol_table ();
int inst_is_breakpoint ();
void k_data_begins_at_point ();
void k_text_begins_at_point ();
void make_memory ();
instruction *r_type_inst ();
int run_spim ();
instruction *set_breakpoint ();
void text_begins_at_point ();
int yylex ();
int yyparse ();


/* Internal functions: */

static mem_addr copy_int_to_stack ();
static mem_addr copy_str_to_stack ();
static void delete_all_breakpoints ();


/* Global Variables: */

int bare_machine = 0;		/* Non-zero => ignore assembler
				   embellishments to bare hardware */

int quiet = 0;			/* Non-zero => no message on traps. */

char *input_file_name;		/* Name of file being parsed */

extern int yylval;		/* Value of token from YYLEX */

int message_out = 0, console_out = 0;

mem_addr program_starting_address = 0;



/* Initialize or reinitialize the state of the machine. */

void
initialize_world (load_trap_handler)
     int load_trap_handler;
{
/*  fprintf(stderr,"in initialize_world()\n"); */
  
  /* Allocate the floating point registers */
  if (FGR == NULL)
    FPR = (double *) malloc (16 * sizeof (double));

  /* Allocate the memory */
/*  fprintf(stderr,"calling make_memory()\n"); */
  make_memory (TEXT_SIZE, DATA_SIZE, STACK_SIZE, K_TEXT_SIZE, K_DATA_SIZE);
  k_text_begins_at_point (K_TEXT_BOT);
  k_data_begins_at_point (K_DATA_BOT);
  data_begins_at_point (DATA_BOT);
  text_begins_at_point (TEXT_BOT);
  initialize_registers ();
  initialize_symbol_table ();
  if (load_trap_handler)
    {
      if(read_assembly_file (DEFAULT_TRAP_HANDLER))
	fatal_error ("Cannot read trap handler");
    }
/*  fprintf(stderr,"calling initialize_scanner()\n"); */
  initialize_scanner (stdin);
/*  fprintf(stderr,"calling delete_all_breakpoints()\n"); */
  delete_all_breakpoints ();
/*  fprintf(stderr,"calling write_output()\n"); */
  write_output (message_out, "SPIM %s\n", SPIM_VERSION);
/*  fprintf(stderr,"exiting initialize_world()\n"); */
}



void initialize_registers ()
{
  bzero (FPR, 16 * sizeof (double));
  FGR = (float *) FPR;
  FWR = (int *) FPR;
  bzero (R, 32 * sizeof (reg_word));
  R[29] = STACK_TOP - BYTES_PER_WORD - 4096; /* Initialize $sp */
  PC = 0;
  Cause = 0;
  EPC = 0;
  Status_Reg = 0;
  BadVAddr = 0;
  Context = 0;
  PRId = 0;
}


/* Read file NAME, which should contain assembly code. Return zero if
   successful and non-zero otherwise. */

int
read_assembly_file (name)
     char *name;
{
  FILE *file = fopen (name, "r");;

  source_file = 1;
  if (file == NULL)
    {
      error ("Cannot open file: `%s'\n", name);
      return (1);
    }
  else
    {
#ifdef mips
#include <sys/exec.h>
      unsigned short magic;

      fread (&magic, sizeof (short), 1, file);
      fclose (file);
      if (magic == MIPSEBMAGIC || magic == MIPSELMAGIC)
	{
	  error ("Source file appears to be executable: %s\n", name);
	  return (1);
	}
      fopen (name, "r");
#endif
      input_file_name = name;
      initialize_scanner (file);
      while (yyparse ()) ;
      fclose (file);
      flush_local_labels ();
      return (0);
    }
}


mem_addr
starting_address ()
{
  if (PC == 0)
    {
      if (program_starting_address)
	return (program_starting_address);
      else
	return (program_starting_address
		= find_symbol_address (DEFAULT_RUN_LOCATION));
    }
  else
    return (PC);
}


/* Initialize the SPIM stack with ARGC, ARGV, and ENVP data. */

void
initialize_run_stack (argc, argv)
     int argc;
     char **argv;
{
  char **p;
  extern char **environ;
  int i, j = 0, env_j;
  mem_addr addrs[10000];

  /* Put strings on stack: */
  for (p = environ; *p != '\0'; p++)
    addrs[j++] = copy_str_to_stack (*p);

  env_j = j;
  for (i = 0; i < argc; i++)
    addrs[j++] = copy_str_to_stack (argv[i]);

  R[29] -= 4;			/* Leave rest of word empty */
  /* Build vectors on stack: */
  for (i = env_j - 1; i >= 0; i--)
    copy_int_to_stack (addrs[i]);
  for (i = j - 1; i >= env_j; i--)
    copy_int_to_stack (addrs[i]);

  R[29] = copy_int_to_stack (argc); /* Leave pointing to argc */
}


static mem_addr
copy_str_to_stack (s)
     char *s;
{
  mem_addr str_start;
  int i = strlen (s);

  while (i >= 0)
    {
      SET_MEM_BYTE (R[29], s[i]);
      R[29] -= 1;
      i -= 1;
    }
  str_start = (mem_addr) R[29] + 1;
  R[29] = R[29] & 0xfffffffc;	/* Round down to word boundary */
  return (str_start);
}


static mem_addr
copy_int_to_stack (n)
     int n;
{
  SET_MEM_WORD (R[29], n);
  R[29] -= BYTES_PER_WORD;
  return ((mem_addr) R[29] + BYTES_PER_WORD);
}


/* Run a program starting at PC for N steps and display each
   instruction before executing if FLAG is non-zero.  If CONTINUE is
   non-zero, then step through a breakpoint.  Return non-zero if
   breakpoint is encountered. */

int
run_program (pc, steps, display, cont_bkpt)
     mem_addr pc;
     int steps, display, cont_bkpt;
{
  if (cont_bkpt && inst_is_breakpoint (pc))
    {
      mem_addr addr = PC;

      delete_breakpoint (addr);
      exception_occurred = 0;
      run_spim (addr, 1, display);
      add_breakpoint (addr);
      steps -= 1;
      pc = PC;
    }

  exception_occurred = 0;
  if (!run_spim (pc, steps, display))
    /* Can't restart program */
    PC = 0;
  if (exception_occurred && Cause == (BKPT_EXCPT << 2))
    return (1);
  else
    return (0);
}


/* Record of where a breakpoint was placed and the instruction previously
   in memory. */

typedef struct bkptrec
{
  mem_addr addr;
  instruction *inst;
  struct bkptrec *next;
} bkpt;


static bkpt *bkpts = NULL;


/* Set a breakpoint at memory location ADDR. */

void
add_breakpoint (addr)
     mem_addr addr;
{
  bkpt *rec = (bkpt *) malloc (sizeof (bkpt));

  rec->next = bkpts;
  rec->addr = addr;

  if ((rec->inst = set_breakpoint (addr)) != NULL)
    bkpts = rec;
  else
    {
      if (exception_occurred)
	error ("Cannot put a breakpoint at address 0x%08x\n", addr);
      else
	error ("Already have a breakpoint at address 0x%08x\n", addr);
      free (rec);
    }
}


/* Delete all breakpoints at memory location ADDR. */

void
delete_breakpoint (addr)
     mem_addr addr;
{
  bkpt *p, *b;
  int deleted_one = 0;

  for (p = NULL, b = bkpts; b != NULL; )
    if (b->addr == addr)
      {
	bkpt *n;

	SET_MEM_INST (addr, b->inst);
	if (p == NULL)
	  bkpts = b->next;
	else
	  p->next = b->next;
	n = b->next;
	free (b);
	b = n;
	deleted_one = 1;
      }
    else
      p = b, b = b->next;
  if (!deleted_one)
    error ("No breakpoint to delete at 0x%08x\n", addr);
}


static void
delete_all_breakpoints ()
{
  bkpt *b, *n;

  for (b = bkpts, n = NULL; b != NULL; b = n)
    {
      n = b->next;
      free (b);
    }
  bkpts = NULL;
}


/* List all breakpoints. */

void
list_breakpoints ()
{
  bkpt *b;

  if (bkpts)
    for (b = bkpts;  b != NULL; b = b->next)
      write_output (message_out, "Breakpoint at 0x%08x\n", b->addr);
  else
    write_output (message_out, "No breakpoints set\n");
}



/* Utility routines */

/* Print the error message then exit. */

/*VARARGS0*/
void
fatal_error (va_alist)
va_dcl
{
  va_list args;
  char *fmt;

  va_start (args);
  fmt = va_arg (args, char *);
#if defined(mips) || defined(__linux__)
  vfprintf (stderr, fmt, args);
#else
  _doprnt (fmt, args, stderr);
#endif
  va_end (args);
  exit (-1);
  /*NOTREACHED*/
}


/* Return the entry in the hash TABLE of length LENGTH with key STRING.
   Return NULL if no such entry exists. */

inst_info *
map_string_to_inst_info (tbl, tbl_len, id)
     register inst_info tbl [];
     int tbl_len;
     register char *id;
{
  register int low = 0;
  register int hi = tbl_len - 1;

  while (low <= hi)
    {
      register int mid = (low + hi) / 2;
      register char *idp = id, *np = tbl[mid].name;

      while (*idp == *np && *idp != '\0') {idp ++; np ++;}

      if (*np == '\0' && *idp == '\0') /* End of both strings */
	return (& tbl[mid]);
      else if (*idp > *np)
	low = mid + 1;
      else
	hi = mid - 1;
    }

  return NULL;
}


/* Return the entry in the hash TABLE of length LENGTH with VALUE1 field NUM.
   Return NULL if no such entry exists. */

inst_info *
map_int_to_inst_info (tbl, tbl_len, num)
     register inst_info tbl [];
     int tbl_len;
     register int num;
{
  register int low = 0;
  register int hi = tbl_len - 1;

  while (low <= hi)
    {
      register int mid = (low + hi) / 2;

      if (tbl[mid].value1 == num)
	return (&tbl[mid]);
      else if (num > tbl[mid].value1)
	low = mid + 1;
      else
	hi = mid - 1;
    }

  return NULL;
}


#ifdef NEED_VSPRINTF
char *
vsprintf (str, fmt, args)
     char *str, *fmt;
     va_list *args;
{
  FILE _strbuf;

  _strbuf._flag = _IOWRT+_IOSTRG;
  _strbuf._ptr = str;
  _strbuf._cnt = 32767;
  _doprnt(fmt, args, &_strbuf);
  putc('\0', &_strbuf);
  return(str);
}
#endif


#ifdef NEED_STRTOL
long
strtol (str, eptr, base)
     char *str, *eptr;
     int base;
{
  long result;

  if (base != 16)
    fatal_error ("SPIM's strtol only works for base 16");
  if (*str == '0' && (*(str + 1) == 'x' || *(str + 1) == 'X'))
    str += 2;
  sscanf (str, "%lx", &result);
  return (result);
}
#endif
